/* ziplist.c */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "ziplist.h"

#define u8 unsigned char
#define u16 unsigned short
#define u32 unsigned long

static int charAt ( const char * string, char search )
	{
		unsigned int i;
		
		for (i=0;i<strlen(string);i++)
			{
				if (string[i]==search)
					return i;
			}
		
		return -1;
	}

static int fsize ( const char * fname )
  {
    FILE * f;
    int s;

    f = fopen (fname,"rb");
    if (f==NULL)
      return -1;

    fseek(f, 0, SEEK_END);
    s=ftell(f);
    fclose(f);

    return s;
  }

static u16 conv_word ( const u8 str[] )
	{
		u16 result;
		
		result = str[0] | str[1]<<8;
		
		return result;
	}

static u32 conv_dword ( const u8 str[] )
	{
		u32 result;
		
		result = conv_word ( &str[0] ) |
		         conv_word ( &str[2] ) << 16;
		
		return result;
	}

int load_ziplist ( const char * fname, ZIPLIST * ziplist )
	{
		u8 * mem;
		int filesize;
		FILE * fh;
		int i;
		int zip_num_entries;
		int strofs;
		int dlistofs, curofs;
		char id[4];
		u8 temp[4];

		if ( fname==NULL || ziplist==NULL )
			return -1;

    memset ( ziplist, 0, sizeof(ZIPLIST) );

		filesize = fsize ( fname );
		if (filesize<0)
			return -1;
			
		mem = malloc ( filesize-16 );
		fh = fopen ( fname, "rb" );
		if (fh==NULL)
			return -1;

		fread ( id, 4, 1, fh );
		if (memcmp ( id, "LST1", 4 ) != 0)
			{
				fclose ( fh );
				free ( mem );
				return -1;
			}
		fread ( temp, 4, 1, fh );
		zip_num_entries = conv_dword ( temp );
		fseek(fh, filesize-8, SEEK_SET);
		fread ( temp, 4, 1, fh );
		ziplist->dirlistsize = conv_dword ( temp );
		fread ( temp, 4, 1, fh );
		dlistofs = conv_dword ( temp );
		fseek(fh, 8, SEEK_SET);
		fread ( mem, filesize-16, 1, fh );
		fclose ( fh );
		
		ziplist->ziplist = (FINFO*)mem;
		ziplist->dirlist = malloc ( ziplist->dirlistsize * sizeof(DINFO) );
		
		for (i=strofs=0;i<zip_num_entries;i++)
			{
				int slen = (int)ziplist->ziplist[i].name;
				ziplist->ziplist[i].name = (char*)&mem[(zip_num_entries*sizeof(FINFO))+strofs];
				strofs += slen+1;
			}
		
		for (i=0, curofs=dlistofs;i<ziplist->dirlistsize;i++)
			{
				int slen = conv_dword ( &mem[curofs] );
				curofs+=4;
				ziplist->dirlist[i].name = (char*)&mem[curofs];
				curofs+=slen+1;
				ziplist->dirlist[i].size = conv_dword ( &mem[curofs] );
				curofs+=4;
				ziplist->dirlist[i].dir = (int*)&mem[curofs];
				curofs+=ziplist->dirlist[i].size*4;
			}

		ziplist->ziplistsize = zip_num_entries;
		
		return 0;
	}

void free_ziplist ( ZIPLIST * ziplist )
	{
		int i;
		
		if (ziplist)
			{
				for (i=0;i<ziplist->ziplistsize;i++)
					{
						if (ziplist->ziplist[i].name)
							{
								free (ziplist->ziplist[i].name);
								ziplist->ziplist[i].name = NULL;
							}
					}
				if (ziplist->ziplist)
					{
						free (ziplist->ziplist);
						ziplist->ziplist = NULL;
					}
				if (ziplist->dirlist)
					{
						free (ziplist->dirlist);
						ziplist->dirlist = NULL;
					}
			}
	}

void dumpdir ( FILE * fh, FINFO * ziplist, int ziplistsize, char * dir, int * dirsize )
	{
		FINFO ** pathlist;
		int pathlistsize;
		int tmp;
		
		printf("DIR: \"%s\"\n", dir);

		(*dirsize)++;

		tmp = strlen(dir);
		fwrite ( &tmp, 4, 1, fh );
		fwrite ( dir, tmp, 1, fh );
		fputc ( '\0', fh );

		pathlist=ziplist_dir(dir, ziplist, ziplistsize, &pathlistsize);
		
		if (pathlist!=NULL)
			{
				int i;

				fwrite ( &pathlistsize, 4, 1, fh );

				for (i=0;i<pathlistsize;i++)
					{
						int tmp = pathlist[i] - ziplist;
						fwrite ( &tmp, 4, 1, fh );
/*						printf("%d ", pathlist[i] - ziplist);
						printf ("\t%s%s %s", pathlist[i]->isdir?"> ":"",
						 &pathlist[i]->name[strlen(dir)],
						 &ziplist[pathlist[i]-ziplist].name[strlen(dir)]);
						if (!pathlist[i]->isdir)
							printf(" (%d bytes)", pathlist[i]->size);
						printf("\n");*/
					}

				for (i=0;i<pathlistsize;i++)
					{
						if (pathlist[i]->isdir)
							{
								char * newpath;

								newpath = malloc ( strlen(dir)+strlen(pathlist[i]->name)+1 );
								strcpy (newpath, dir);
								strcat (newpath, &pathlist[i]->name[strlen(dir)]);
								dumpdir ( fh, ziplist, ziplistsize, newpath, dirsize );
								free ( newpath );
							}
					}
				free ( pathlist );
			}
	}

int dumpdirs ( FILE * fh, FINFO * ziplist, int ziplistsize )
	{
		int dirsize = 0;
		dumpdir ( fh, ziplist, ziplistsize, "", &dirsize );
		
		return dirsize;
	}

int build_ziplist ( const char * fname, ZIPLIST * ziplist )
	{
    unzFile uf = NULL;
    unz_global_info gi;
    unz_file_info file_info;
    int i;
    int zip_num_entries;
    char * filename;

    if (fname==NULL || ziplist==NULL)
    	return -1;

    memset ( ziplist, 0, sizeof(ZIPLIST) );
    
    uf = unzOpen ( fname );
    if (uf==NULL)
      {
        printf("Couldn't open file\n");
        return -1;
      }

    if (unzGetGlobalInfo ( uf, &gi ) != UNZ_OK)
      {
        unzClose ( uf );
        return -1;
      }

    zip_num_entries = gi.number_entry;

    ziplist->ziplist = malloc ( zip_num_entries*sizeof(FINFO) );
    memset ( ziplist->ziplist, 0, sizeof(FINFO)*zip_num_entries );

    filename = malloc(65536);
    for (i=0;i<zip_num_entries;i++)
      {
        unzGetCurrentFileInfo ( uf, &file_info, filename, 65536, NULL, 0, NULL, 0 );
        ziplist->ziplist[i].name = malloc ( file_info.size_filename+1 );
        strcpy ( ziplist->ziplist[i].name, filename );
        ziplist->ziplist[i].size = file_info.uncompressed_size;
        ziplist->ziplist[i].isdir = (file_info.flag&2)==0;
        unzGoToNextFile(uf);
      }
    free (filename);

    unzClose(uf);
    
    ziplist->ziplistsize = zip_num_entries;

    return 0;
	}

int save_ziplist ( const char * fname, ZIPLIST * ziplist )
	{
		int i;
		int dlistsize;
		int dlistofs;
		FILE * fhandle;

		if ( fname==NULL || ziplist==NULL )
			return -1;

    fhandle = fopen ( fname, "wb" );
    if (fhandle==NULL)
    	return -1;

    fwrite ( "LST1", 1, 4, fhandle );
    fwrite ( &ziplist->ziplistsize, 1, 4, fhandle );

    for (i=0;i<ziplist->ziplistsize;i++)
    	{
	    	int l = strlen ( ziplist->ziplist[i].name );
	    	fwrite ( &ziplist->ziplist[i].size, 4, 1, fhandle );
	    	fwrite ( &ziplist->ziplist[i].isdir, 4, 1, fhandle );
	    	fwrite ( &l, 4, 1, fhandle );
	    	fwrite ( &ziplist->ziplist[i].userdata, 4, 1, fhandle );
    	}

    for (i=0;i<ziplist->ziplistsize;i++)
    	{
	    	fwrite ( ziplist->ziplist[i].name, strlen(ziplist->ziplist[i].name), 1, fhandle );
	    	fputc ( '\0', fhandle );
    	}

    dlistofs = ftell(fhandle)-8;
		dlistsize = dumpdirs ( fhandle, ziplist->ziplist, ziplist->ziplistsize );
		fwrite ( &dlistsize, 4, 1, fhandle );
		fwrite ( &dlistofs, 4, 1, fhandle );

    fclose (fhandle);

    return 0;
	}

int ziplist_getdir ( const char * pathname, FINFO * ziplist, int ziplistsize, FINFO ** dest )
	{
		int i;
		int pathlistsize;

    for (i=pathlistsize=0;i<ziplistsize;i++)
    	{
	    	int charofs;
	    	
	    	charofs = charAt(ziplist[i].name,'/');

	    	if ( (strncmp(pathname, ziplist[i].name, strlen(pathname))==0) &&
	    	     (strlen(ziplist[i].name)>strlen(pathname))  )
	    		{
		    		int pathnamelen = strlen(pathname);
		    		int charofs = charAt(&ziplist[i].name[pathnamelen], '/');

		    		if (charofs<(int)0 || charofs==(int)strlen(&ziplist[i].name[pathnamelen])-1)
		    			{
	    					if (dest!=NULL)
	    						dest[pathlistsize]=&ziplist[i];
/*	    					printf ("%s\n", ziplist[i].name);*/
	    					pathlistsize++;
    					}
    			}
    	}
    return pathlistsize;
   }

static void lsort ( FINFO ** dir, int start, int end, int dirsize )
	{
		int i;

		if (dirsize<=1 || start<0 || end<0 || start>=dirsize || end>=dirsize || start>end)
			return;

		for (i=start;i<=end-1;i++)
			{
				int j;
				
				for (j=i+1;j<=end;j++)
					{
						if (strcmp(dir[i]->name, dir[j]->name)>0)
							{
								FINFO * tmp;
								
								tmp = dir[i];
								dir[i] = dir[j];
								dir[j] = tmp;
							}
					}
			}
	}

FINFO ** ziplist_directory ( const char * pathname, ZIPLIST * ziplist, int * dirsize )
	{
		FINFO ** dest;
		FINFO * zlist;
		DINFO * dlist;
		int i;

		if (pathname==NULL || ziplist==NULL || dirsize==NULL)
			return NULL;
		
		if (ziplist->ziplist==NULL || ziplist->dirlist==NULL)
			return NULL;

		zlist = ziplist->ziplist;
		dlist = ziplist->dirlist;
			
		for (i=0;i<ziplist->dirlistsize;i++)
			{
				if (strcmp(dlist[i].name, pathname)==0)
					{
						int j;
						int s;

						s = dlist[i].size;
						*dirsize = s;
						
						dest = malloc ( s * sizeof(FINFO*) );
						memset ( dest, 0, sizeof(FINFO*)*s );
						
						if (dest==NULL)
							return NULL;

						for (j=0;j<s;j++)
							{
								int ofs = conv_dword ( (u8*)&dlist[i].dir[j] );
								dest[j] = &zlist[ofs];
							}

						return dest;
					}
			}

		return NULL;
	}

void sortdir ( FINFO ** dest, int dirsize )
	{
		int i,numfiles,numdirs;

		if (dirsize<=0 || dest==NULL)
			return;

		for (i=numfiles=numdirs=0;i<dirsize;i++)
			{
				if (dest[i]->isdir)
					numdirs++;
				else
					numfiles++;
			}
		
		if (numdirs!=0 && numfiles!=0)
			{
				/* Separate files from dirs */
				for (i=0;i<dirsize-1;i++)
					{
						int j;
						
						for (j=i+1;j<dirsize;j++)
							{
								if (!dest[i]->isdir && dest[j]->isdir)
									{
										FINFO * tmp;
										tmp = dest[i];
										dest[i] = dest[j];
										dest[j] = tmp;
									}
							}
					}
			}

		lsort ( dest, 0, numdirs-1, dirsize );
		lsort ( dest, numdirs, dirsize-1, dirsize );
	}
	
FINFO ** ziplist_dir ( const char * pathname, FINFO * ziplist, int ziplistsize, int * dirsize )
	{
		FINFO ** dest = NULL;
		
		if (pathname==NULL || ziplist==NULL || ziplistsize<0 || dest==NULL)
			return NULL;

		*dirsize = ziplist_getdir ( pathname, ziplist, ziplistsize, NULL );
		if (*dirsize<0)
			return NULL;

		dest = calloc ( sizeof(FINFO*), *dirsize );

		ziplist_getdir ( pathname, ziplist, ziplistsize, dest );
		
		sortdir ( dest, *dirsize );

		return dest;
	}
